#!/usr/bin/perl

# domo.pl

# Description: This script listens on a specified serial port for the
# data received from UHF remote sensors.
#
# Options:
# usage: ./domo.pl [-D] [-v] [-p port] [-h]
#     -h        : this usage information
#     -D        : do not daemonize (stay in foreground)
#     -v        : verbose debugging output
#     -l        : log output
#     -p port   : serial port
#     -t format : format of timestamp (strftime)

our $VERSION = "0.8";

# pragmas
use warnings;
use strict;

# modules
use lib qw( /usr/local/perl );
use Getopt::Std;
use Errno qw(EINTR);
use POSIX qw( strftime setsid );
use Cwd qw(realpath);
use File::Basename;

# constants
use constant TRUE  => 1;
use constant FALSE => 0;

use constant PORT => '/dev/ttyS0';
use constant DOMOPATH => '/var/lib/domo';
use constant DOMO => '/var/lib/domo/domo.dat';
use constant DOMOSTAT => '/var/lib/domo/domo.stat';
use constant DOMOTMP => '/var/lib/domo/domo.tmp';
use constant DOMOLOG => '/var/log/domo';
use constant TIMEFORMAT => '%d-%m-%y %T';
use constant PROCESS   => 300;	# process each 300 seconds

use constant SYNC   => 1;
use constant TOFROM => 2;
use constant SEQLEN => 3;
use constant DATA   => 4;
use constant CHKSUM => 5;

# Stats
my $charLost = 0;
my $pingLost = 0;
my $checkErr = 0;
my $charRx = 0;
my $charTx = 0;
my $nbPings = 0;
my $nbTxFrames = 0;
my $nbRxFrames = 0;
my $unknown = 0;

my $verbose    = FALSE;
my $daemon     = TRUE;
my $run        = TRUE;
my $log        = FALSE;
my $tty        = PORT;
my $timeformat = TIMEFORMAT;

$SIG{TERM} = $SIG{INT}  = sub { print "Aborted\n"; $run = FALSE; };

parseArgs();

$SIG{'USR1'} = 'USR1_handler';
printf strftime($timeformat, localtime) . " Server started\n";

my @recip = (1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1);
my @mods;
my @plugins;
# my @preWrite;
my $fullpath = realpath($0);
my $path = dirname($fullpath);

print "changing directory to $path\n";
chdir($path);

mkdir DOMOPATH if (! -d DOMOPATH);

&loadModules("modules", \@mods);
&loadModules("plugins", \@plugins);
# &loadModules("preWrite", \@preWrite);

my $d;
my @data;
my @data_p;
my @pingReq;

my $byte;
my $idx;
my $chck;
my $seq = 0;
my $mode = SYNC;
my $delay = time + 3;
my $process = &nextProcess;

my $pingAddr = 0;
$pingReq[0] = -1;

my %mesures;

my $port = initSerial($tty);

if (&checkChain($port, $d) == 0) {
	close($port);
	printf STDERR "Error in RS232 chain\n";
	exit 1;
}

printf strftime($timeformat, localtime) . " Server ready\n";

$daemon && daemonize();


while ($run) {
	if (&getFrame($port)) {
		if (!compareArrays(\@data, \@data_p)) {
			my $head = '';
			$head .= sprintf strftime($timeformat, localtime) . " len=%-2d ", scalar(@data);
			$head .= sprintf "%02x", $_ foreach (@data);
			$head .= sprintf " :";
			if (!&processFrame($port, $head, $d, @data)) {
				print "$head Did not process frame\n" if ($verbose);
				++$unknown;
			}
			@data_p = @data;

		}
	}
	my $t = time;
	if ($t > $delay) {
		# $delay += 10;
		$delay += 2;
		if ($pingReq[$pingAddr] != -1 && $recip[$pingAddr]) {
			dd("Ping to $pingAddr lost");
			print "Ping to $pingAddr lost\n" if ($verbose);
			++$pingLost;
		}

		$pingAddr++;
		$pingAddr = 1 if ($pingAddr > 15);
		&ping($port, $pingAddr) if ($recip[$pingAddr]);
	}
	if ($t > $process) {
		&writeData();
		$process = &nextProcess;
	}

	&getCommand($port, $t);
}

close($port);

exit(0);

sub USR1_handler {
	if ($log) {
		close LOG;
		open LOG,">>".DOMOLOG;
		select(LOG);
		$| = 1;
		print("USR1 Interrupt: LOG validated!\n");
	}
}

my @bytearray;

sub getChar {
	my ($p) = @_;

	# printf "array len = %d\n", scalar(@bytearray);
	return shift(@bytearray) if (scalar(@bytearray) > 0);

	if (isChar($p)) {
		my $bufrcv;
		my $nb = sysread($p, $bufrcv, 50);
		# print "nb=$nb\n";
		return -1 if ($nb <= 0);

		$charRx += $nb;

		push @bytearray, unpack "C*", $bufrcv;
		return shift(@bytearray);
	}
	return -1;
}

sub getFrame {
	my ($p) = @_;

	my $frame = 0;

	while ((my $byte = getChar($p)) != -1) {
		# printf("mode=$mode byte=%02x len=%d\n", $byte, scalar(@data));
		if ($mode == SYNC) {
			$idx = 0;
			$chck = 0;
			$#data = -1;
			$mode = TOFROM if ($byte == 2);
			$charLost++ if ($byte != 2);
		}
		elsif ($mode == TOFROM) {
			$d->{'from'} = $byte >> 4;
			$d->{'to'}   = $byte & 0xf;
			$chck += $byte;
			$mode = SEQLEN;
		}
		elsif ($mode == SEQLEN) {
			$d->{'seq'} = $byte >> 4;
			$d->{'len'} = ($byte & 0xf) + 1;
			$chck += $byte;
			$mode = DATA;
		}
		elsif ($mode == DATA) {
			$data[$idx++] = $byte;
			$chck += $byte;
			$mode = CHKSUM if ($idx == $d->{'len'});
		}
		elsif ($mode == CHKSUM) {
			$chck += $byte;
			$chck &= 0xff;
# print "received frame of $idx bytes\n" if ($verbose);
			$mode = SYNC;
			$checkErr++ if ($chck != 0);
			return 1 if ($chck == 0);
		}
	}

	return 0;
}

sub clearStats {
	$charLost = 0;
	$pingLost = 0;
	$checkErr = 0;
	$charRx = 0;
	$charTx = 0;
	$nbPings = 0;
	$nbTxFrames = 0;
	$nbRxFrames = 0;
	$unknown = 0;
}

sub checkChain {
	my ($p) = @_;

	my @send = (0x02, 0x00, 0x00, 0xaa, 0x00);
	$send[2] = &sequence;
	$send[4] = -($send[1] + $send[2] + $send[3]) & 0xff;

	my $buf = pack "C5", @send;
	my $nb = syswrite($p, $buf, 5);
	$charTx += 5;
	++$nbTxFrames;

	while (&getFrame($p)) {
		return 1 if ($d->{'from'} == 0 && $d->{'to'} == 0 && $data[0] == 0xaa);
	}

	return 0;
}

sub nextProcess {
	my $t = time;
	my $o = PROCESS - 10 - ($t % PROCESS);
	$o += PROCESS if ($o <= 60);	# minimum 60 seconds period
	return $t + $o;
}

sub compareArrays {
	my ($first, $second) = @_;
	no warnings;  # silence spurious -w undef complaints
	return 0 unless @$first == @$second;
	for (my $i = 0; $i < @$first; $i++) {
		return 0 if $first->[$i] ne $second->[$i];
	}
	return 1;
}

sub loadModules {
	my ($dir, $add_mods) = @_;

	print "Begin loading $dir\n";

	my @filelist;

	opendir(D, $dir) or die $!;
	push @filelist, $_ while(<$dir/*.pm>);
	closedir D;
	
	foreach my $module (@filelist) {
		$module =~ s!^(?:.*/)?(.+?)(?:\.[^.]*)?$!$1!;
		eval "use ${dir}::$module";
		if ($@) {
			print "*** Loading module ${dir}::$module error\n$@\n";
		}
		else {
			# print "module $module installed\n";
			my $p = $module->new();
			$p->verbose($verbose);
			push @$add_mods, $p;
		}
	}
	print "End loading $dir\n";
}

sub processFrame
{
	my ($p, $head, $s, @data) = @_;

	if ($verbose) {
	# if ($data[0] == 0) {
	#	print strftime $timeformat, localtime();
	#	printf " from=%d to=%d seq=%-2d len=%-2d ",
	#			$s->{'from'}, $s->{'to'}, $s->{'seq'}, $s->{'len'};
	#	printf "%02x ", $_ foreach (@data);
	#	print "\n";
	# }
	}

	++$nbRxFrames;
	my $ok = 0;

	# ping 'from' is me, the packet traveled through the chain
	# and loops back to me.
	# the recipient does not exist. Remove it from list.
	if ($s->{'from'} == 0 && $data[0] == 0x01) {
		my $addr = $s->{'to'};
		$recip[$addr] = 0;
		print "Address $addr does not exist. Removed.\n" if ($verbose);
		return 1;
	}

	return if ($s->{'to'} != 0);

	if ($data[0] == 0x01) {
		# ping
		$ok = processPing($head, $p, $s->{'from'}, @data) if ($data[0] == 0x01);
	}
	elsif ($data[0] == 0x02) {
		# ident
		$ok = processIdent($head, $p, $s->{'from'}, @data) if ($data[0] == 0x02);
	}
	else {
		# other frames

		my ($sc, $mn) = localtime(time);
		my $min = int($mn / 5);

		$ok |= $_->process($head, $timeformat, $min, @data) foreach (@mods);
	}

	return $ok;
}

sub getCommand
{
	my ($p, $t) = @_;

	my $ok = 0;

	foreach my $mod (@mods) {
		my @data;

		if ($mod->command($t, \@data)) {

			my $len = scalar(@data);
#print "len=$len\n";
			$data[0] = 0x02;
			$data[2] = &sequence | ($len-4);

			my $addr = $data[1];
			my $cmd  = $data[3];
			print strftime($timeformat, localtime) . " -> command $cmd to $addr\n" if ($verbose);

			my $chk = 0;
			my $i;
# print "02 ";
			for ($i = 1 ; $i < $len ; $i++) {
# printf "%02x ", $data[$i];
				$chk += $data[$i];
			}
			$data[$i] = -($chk) & 0xff;
# printf "%02x\n", $data[$i];

			my $buf = pack "C*", @data;
			my $nb = syswrite($p, $buf, $len+3);
			$charTx += $len+3;
			++$nbTxFrames;
		}
	}
}

sub processPing
{
	my ($head, $p, $from, @data) = @_;

	printf "$head Ping to $from request:%02x answered:%02x\n", $pingReq[$from], $data[1] if ($verbose);
	if ($pingReq[$from] == $data[1] && $recip[$from]) {
		$pingReq[$from] = -1;
		&ident($p, $from) if ($recip[$from] == 1);
		return 1;
	}
	return 0;
}

sub processIdent
{
	my ($head, $p, $from, @data) = @_;

	shift(@data);		# remove command 0x02
	my $len = scalar(@data);
	my $str = pack "C$len", @data;
	printf "$head Ident from $from (len=$len): <%s>\n", $str if ($verbose);
	$recip[$from] = 2;
	return 1;
}

# return sequence nunmber in MSB
sub sequence
{
	my $val = $seq++;
	$seq = 1 if ($seq == 16);
	$val <<= 4;
	return $val;
}

# parameter = to address
sub ident
{
	my ($p, $addr) = @_;

	print strftime($timeformat, localtime) . " -> ident to $addr\n" if ($verbose);

	my $len = 0;			# 1 byte - 1

	my @data = (0x02, 0x00, 0x00, 0x00, 0x00);
	$data[1] = $addr;
	$data[2] = &sequence | $len;
	$data[3] = 2;			# ident command
	$data[4] = -($data[1] + $data[2] + $data[3]) & 0xff;

	my $buf = pack "C5", @data;
	dd("ident\n");
	my $nb = syswrite($p, $buf, 5);
	$charTx += 5;
	++$nbTxFrames;
	# my $nb = print $p "$buf";
	# print("end ident nb = $nb\n");
	dd("end ident nb = $nb\n");
}

# parameter = to address
sub ping
{
	my ($p, $addr) = @_;

	# print strftime($timeformat, localtime) . " -> ping to $addr\n" if ($verbose);

	$pingReq[$addr] = int(rand(256));

	my $len = 1;			# 2 bytes - 1

	my @data = (0x02, 0x00, 0x00, 0x01, 0x00, 0x00);
	$data[1] = $addr;
	$data[2] = &sequence | $len;
	$data[3] = 1;			# ping command
	$data[4] = $pingReq[$addr];	# payload
	$data[5] = -($data[1] + $data[2] + $data[3] + $data[4]) & 0xff;

	my $buf = pack "C6", @data;
	dd("ping\n");
	my $nb = syswrite($p, $buf, 6);
	$charTx += 6;
	++$nbTxFrames;
	# my $nb = print $p "$buf";
	# print("end ping nb = $nb\n");
	dd("end ping nb = $nb\n");

	++$nbPings;
}

sub writeData
{
	my ($add, $val, $type, $v);

	my $t = time;

	my ($sc, $mn) = localtime($t);
	my $min = int($mn / 5);

	my $d = $t . " " . &strftime($timeformat, localtime($t)) . "\n";


	my $str = '';
	my @prelines;

	foreach (@mods) {
		my $s = $_->get($min);
		push @prelines, $s;
		$str .= $s;
	}

	# Call prewrite modules
	# foreach my $pr (@preWrite) {
		# $pr->begin($min);
		# $pr->process($_) foreach (@prelines);
		# $pr->end();
	# }

#	print "\n$str\n";
	open  TMP, ">".DOMOTMP;
	print TMP $d.$str;
	close TMP;

	rename DOMOTMP, DOMO;

	# Call postwrite modules 
	foreach my $po (@plugins) {
		printf strftime($timeformat, localtime) . " starting " . $po->name . "\n" if ($verbose);
		$po->process($timeformat, $min, $str);
	}
	printf strftime($timeformat, localtime) . " end plugins\n" if ($verbose);

	$str = $d;
	$str .= sprintf("Sent     characters: %d\n", $charTx);
	$str .= sprintf("Received characters: %d\n", $charRx);
	$str .= sprintf("Spurious characters: %d\n", $charLost);
	$str .= sprintf("Sent     frames    : %d\n", $nbTxFrames);
	$str .= sprintf("Received frames    : %d\n", $nbRxFrames);
	$str .= sprintf("Checksum errors    : %d\n", $checkErr);
	$str .= sprintf("RX data  errors    : %d\n", $unknown);
	$str .= sprintf("Sent     pings     : %d\n", $nbPings);
	$str .= sprintf("Lost     pings     : %d\n", $pingLost);

	open  TMP, ">".DOMOSTAT;
	print TMP $str;
	close TMP;

	rename DOMOTMP, DOMOSTAT;

	print $str if ($verbose);
}

sub dd {
	my ($s) = @_;

	my $timeformat = '%d-%b-%y %T ';
	my $t = strftime $timeformat, localtime();
	# open DEBUG, ">> /var/log/domo_debug";
	# print DEBUG $t.$s;
	# close DEBUG;
}

# parameter = fatal message
sub fatal
{
	my ($msg) = @_;
	
	print STDERR $msg, "\n";
	
	exit(1);
}

sub parseArgs
{
	my %opts;
	
	&getopts('Dlvhp:', \%opts);
	
	$opts{'D'} and do {
		$daemon = FALSE;
	};
	$opts{'l'} and do {
		$verbose  = TRUE;
		$log = TRUE;
		open LOG,">>".DOMOLOG;
		select(LOG);
		$| = 1;
	};
	$opts{'v'} and do {
		$verbose  = TRUE;
	};
	$opts{'p'} and do {
		$tty = $opts{'p'};
	};
	$opts{'t'} and do {
		$timeformat = $opts{'t'};
	};

	if ($opts{'h'}) {
		die "usage: $0 [-D] [-v] [-l] [-p port] [-h]\n",
		"Version: $VERSION\n",
		"    -h                : this usage information\n",
		"    -D                : do not daemonize (stay in foreground)\n",
		"    -v                : verbose debugging output\n",
		"    -l                : log output\n",
		"    -p serial_device  : serial port device\n",
		"    -t time_format    : time format for logging\n";
	}
}

sub initSerial
{
	my ($ref) = @_;
	my $sd;
	
	open($sd, '+<', $tty) ||
		fatal("can't open serial port: " . $tty . ":$!");
	
	# set the parameters
	system("/bin/stty 4800 cs8 cread ignpar -crtscts -echo raw < " . $tty) &&
		fatal("unable to set serial port parameters: $!");
	
	return $sd;
}

sub isChar
{
	my ($port) = @_;
	
	my ($rin, $ein);
	
	$rin = $ein = '';
	vec($rin, fileno($port), 1) = 1;
	$ein = $rin;

	return undef if (select($rin, undef, $ein, 1) == 0);
	return undef if ( $!{EINTR} );
	return TRUE;
}

sub daemonize
    # Turns the process into a daemon.
    # Parameters: None
    # Returns: void
{
    my $pidfile = "/var/run/domo.pl.pid";

    # chdir '/'                 or die "Can't chdir to /: $!";
    open STDIN, '/dev/null'   or die "Can't read /dev/null: $!";
    open STDOUT, '>/dev/null' or die "Can't write to /dev/null: $!" if ($log == FALSE);
    defined(my $pid = fork)   or die "Can't fork: $!";
    if ($pid) {
        open PIDFILE, ">$pidfile" or die "cannot open $pidfile: $!\n";
        print PIDFILE $pid;
        close PIDFILE;
        exit 0;
    }
    setsid                    or die "Can't start a new session: $!";
    open STDERR, '>&STDOUT'   or die "Can't dup stdout: $!";
}
